from __future__ import annotations

import argparse
import contextlib
import logging
import logging.config
import subprocess
import sys
from typing import Literal

import npc_lims
import psutil

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": True,
    }
)

logger = logging.getLogger()

MEMORY_PER_EPHYS_SESSION = 2 * 1024**3
"""Conservative estimate of a whole ephys session in memory."""

MEMORY_PER_TRAINING_SESSION = 0.2 * 1024**3

DEFAULT_CONTAINER_MEMORY = 7 * 1024**3
"""Default available memory in codeocean capsule or github action runner, in case
we can't query it.

https://docs.github.com/en/actions/using-github-hosted-runners/about-github-hosted-runners/about-github-hosted-runners#supported-runners-and-hardware-resources
"""


def get_max_workers(session_type: Literal["ephys", "training", "all"] = "ephys") -> int:
    if session_type == "training":
        memory_per_session = MEMORY_PER_TRAINING_SESSION
    else:
        memory_per_session = MEMORY_PER_EPHYS_SESSION

    return int(
        min(
            psutil.cpu_count(),
            get_available_memory() * 0.8 // memory_per_session,
        )
    )


def get_available_memory() -> int:
    """Assumes linux means containerized - get cgroups memory if possible."""
    if sys.platform == "linux":
        with contextlib.suppress(ValueError):
            return get_available_container_memory()
        return DEFAULT_CONTAINER_MEMORY
    return psutil.virtual_memory().available


def get_available_container_memory() -> int:
    """Available memory in the container, in bytes.

    `psutil.virtual_memory()` gives system memory, not container memory.
    In a github action or codeocean capsule, psutil will overestimate the available memory.
    """
    if sys.platform != "linux":
        raise NotImplementedError("Only implemented for linux")

    def _format(out: bytes) -> int:
        return int(out.decode().strip("\n"))

    def _run(cmd: list[str]) -> bytes:
        return subprocess.run(cmd, capture_output=True).stdout

    limit = _format(_run(["cat", "/sys/fs/cgroup/memory/memory.limit_in_bytes"]))
    usage = _format(_run(["cat", "/sys/fs/cgroup/memory/memory.usage_in_bytes"]))
    return limit - usage


def setup() -> dict[
    Literal["session_type", "skip_existing", "version", "parallel"],
    str | bool | None,
]:
    args = parse_args()
    kwargs = vars(args)
    logging.basicConfig(
        level=kwargs.pop("log_level"),
        format="%(asctime)s | %(name)s | %(levelname)s | %(message)s",
        datefmt="%H:%M:%S",
        stream=sys.stdout,
    )
    logger.info(f"Using parsed args {args}")
    return kwargs  # type: ignore[return-value]


def parse_args() -> argparse.Namespace:
    """
    >>> parse_args()
    Namespace(session_type='ephys', skip_existing=True, version=None, parallel=False, log_level='INFO', max_workers=None)
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--session-type",
        choices=("training", "ephys", "all"),
        default="ephys",
        help="Type of sessions to process and write to cache (default 'ephys')",
    )
    parser.add_argument(
        "--overwrite",
        dest="skip_existing",
        action="store_false",
        help="Flag to skip existing files in cache",
    )
    parser.add_argument(
        "--version",
        type=str,
        metavar="v?.?.?",
        help="Subfolder that cached files are written to - default uses the current version of npc_sessions",
    )
    parser.add_argument(
        "--parallel",
        action="store_true",
        help="Flag to run in parallel",
    )
    parser.add_argument(
        "--log-level",
        default="INFO",
        type=str,
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        help="Level for the root logger (default INFO)",
    )
    parser.add_argument(
        "--max-workers",
        type=int,
        metavar="N",
        help="If --parallel is used, override the number of workers in ProcessPool - default calculates appropriate number based on available memory",
    )

    return parser.parse_args()


def get_session_infos(
    session_type: Literal["training", "ephys", "all"]
) -> tuple[npc_lims.SessionInfo, ...]:
    if session_type not in ("training", "ephys", "all"):
        raise ValueError(f"{session_type=} must be one of 'training', 'ephys', 'all'")
    if session_type == "all":
        session_infos = npc_lims.get_session_info(issues=[])
    else:
        session_infos = tuple(
            s
            for s in npc_lims.get_session_info(issues=[])
            if s.is_ephys == is_ephys(session_type=session_type)
        )
    return session_infos


def is_ephys(session_type: Literal["training", "ephys", "all"]) -> bool | None:
    """
    >>> is_ephys("training")
    False
    >>> is_ephys("ephys")
    True
    >>> is_ephys("all")


    """
    if session_type not in ("training", "ephys", "all"):
        raise ValueError(f"{session_type=} must be one of 'training', 'ephys', 'all'")
    if session_type == "all":
        return None
    return session_type == "ephys"


if __name__ == "__main__":
    import doctest

    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



import datetime
from typing import TYPE_CHECKING, Literal

import numpy as np
from matplotlib import pyplot as plt

import npc_sessions.utils as utils

if TYPE_CHECKING:
    import npc_sessions


def get_audio_latencies(
    session: "npc_sessions.DynamicRoutingSession", stim_type: Literal["task", "mapping"]
):
    if "task" in stim_type.lower() or "behavior" in stim_type.lower():
        sel_stim = next(s for s in session.stim_paths if "DynamicRouting" in s.stem)
    elif "rf" in stim_type.lower() or "mapping" in stim_type.lower():
        sel_stim = next(s for s in session.stim_paths if "RFMapping" in s.stem)

    # get nidaq latencies
    lat_nidaq_env = utils.get_stim_latencies_from_nidaq_recording(
        sel_stim,
        session.sync_path,
        session.ephys_recording_dirs,
        "audio",
        use_envelope=True,
    )

    lat_nidaq_sig = utils.get_stim_latencies_from_nidaq_recording(
        sel_stim,
        session.sync_path,
        session.ephys_recording_dirs,
        "audio",
        use_envelope=False,
    )

    lat_nidaq_sig_list = [y.latency for y in lat_nidaq_sig if y is not None]
    lat_nidaq_env_list = [y.latency for y in lat_nidaq_env if y is not None]

    sound_type_list = [y.name for y in lat_nidaq_sig if y is not None]

    audtridx = []
    for ii, lat in enumerate(lat_nidaq_sig):
        if lat is not None:
            audtridx.append(ii)

    # get sync latencies
    if session.sync_data.start_time.date() >= datetime.date(year=2023, month=8, day=31):
        sync_sound_on = True
    else:
        sync_sound_on = False

    if sync_sound_on:
        sync_line = utils.get_sync_line_for_stim_onset(
            "audio", session.sync_data.start_time.date()
        )

        lat_sync = utils.get_stim_latencies_from_sync(
            sel_stim, session.sync_path, "audio", sync_line
        )

        lat_sync_list = [y.latency for y in lat_sync if y is not None]
        audtridx_sync = []
        for ii, lat in enumerate(lat_sync):
            if lat is not None:
                audtridx_sync.append(ii)

    latency_info = {
        "nidaq_signal": lat_nidaq_sig_list,
        "nidaq_envelope": lat_nidaq_env_list,
        "aud_trial_idx": audtridx,
        "sync_sound_on": sync_sound_on,
        "sound_type": sound_type_list,
    }

    if sync_sound_on:
        latency_info["sync"] = lat_sync_list
        latency_info["aud_trial_idx_sync"] = audtridx_sync

    return latency_info


def plot_audio_latencies(session: "npc_sessions.DynamicRoutingSession", latency_info):
    latency_flags = []
    if np.sum(np.array(latency_info["nidaq_signal"]) < 0) > 0:
        latency_flags.append("signal xcorr")
    if np.sum(np.array(latency_info["nidaq_envelope"]) < 0) > 0:
        latency_flags.append("envelope xcorr")
    if latency_info["sync_sound_on"] is True:
        if np.sum(np.array(latency_info["sync"]) < 0) > 0:
            latency_flags.append("sync")

    xbins = np.arange(-0.15, 0.15, 0.001)
    fig, ax = plt.subplots(2, 1)
    ax[0].plot(
        latency_info["aud_trial_idx"], latency_info["nidaq_signal"], ".", alpha=0.5
    )
    ax[0].plot(
        latency_info["aud_trial_idx"], latency_info["nidaq_envelope"], ".", alpha=0.5
    )

    ax[0].set_xlabel("trial number")
    ax[0].set_ylabel("audio latency (s)")

    ax[1].hist(latency_info["nidaq_signal"], bins=xbins, alpha=0.5)
    ax[1].hist(latency_info["nidaq_envelope"], bins=xbins, alpha=0.5)

    ax[1].set_xlabel("audio latency (s)")
    ax[1].set_ylabel("trial count")
    ax[1].legend(["signal", "envelope"])

    if latency_info["sync_sound_on"] is True:
        ax[0].plot(
            latency_info["aud_trial_idx_sync"], latency_info["sync"], ".", alpha=0.5
        )
        ax[1].hist(latency_info["sync"], bins=xbins, alpha=0.5)
        ax[1].legend(["signal", "envelope", "sync"])

    if session.task_version and "templeton" in session.task_version:
        figtitle = "Audio latency by alignment method: " + session.id + " (templeton)"
    else:
        figtitle = "Audio latency by alignment method: " + session.id + " (DR)"

    latency_warning = "Warning! negative latencies in: "
    if len(latency_flags) > 0:
        for flag in latency_flags:
            latency_warning = latency_warning + flag + ", "
        ax[0].set_title(latency_warning[:-2])

    fig.suptitle(figtitle)
    fig.tight_layout()

    return fig


def plot_tone_vs_AMnoise(session: "npc_sessions.DynamicRoutingSession", latency_info):
    # compare tones vs. AM noise
    tone_idx = np.asarray(latency_info["sound_type"]) == "tone"
    AMnoise_idx = np.asarray(latency_info["sound_type"]) == "AM_noise"

    aud_trial_idx = np.asarray(latency_info["aud_trial_idx"])
    nidaq_signal = np.asarray(latency_info["nidaq_signal"])
    nidaq_envelope = np.asarray(latency_info["nidaq_envelope"])
    if latency_info["sync_sound_on"] is True:
        sync = np.asarray(latency_info["sync"])
        aud_trial_idx_sync = np.asarray(latency_info["aud_trial_idx_sync"])

    xbins = np.arange(-0.15, 0.15, 0.001)
    fig, ax = plt.subplots(2, 2, figsize=(8, 5))

    idxs = [tone_idx, AMnoise_idx]
    idx_labels = ["tone", "AM noise"]

    for xx, idx in enumerate(idxs):
        latency_flags = []
        if np.sum(np.array(nidaq_signal[idx]) < 0) > 0:
            latency_flags.append("signal xcorr")
        if np.sum(np.array(nidaq_envelope[idx]) < 0) > 0:
            latency_flags.append("envelope xcorr")
        if latency_info["sync_sound_on"] is True:
            if np.sum(np.array(sync[idx]) < 0) > 0:
                latency_flags.append("sync")

        ax[0, xx].plot(aud_trial_idx[idx], nidaq_signal[idx], ".", alpha=0.5)
        ax[0, xx].plot(aud_trial_idx[idx], nidaq_envelope[idx], ".", alpha=0.5)

        ax[0, xx].set_xlabel("trial number")
        ax[0, xx].set_ylabel("audio latency (s)")

        ax[1, xx].hist(nidaq_signal[idx], bins=xbins, alpha=0.5)
        ax[1, xx].hist(nidaq_envelope[idx], bins=xbins, alpha=0.5)

        ax[1, xx].set_xlabel("audio latency (s)")
        ax[1, xx].set_ylabel("trial count")
        ax[1, xx].legend(["signal", "envelope"])

        if latency_info["sync_sound_on"] is True:
            ax[0, xx].plot(aud_trial_idx_sync[idx], sync[idx], ".", alpha=0.5)
            ax[1, xx].hist(sync[idx], bins=xbins, alpha=0.5)
            ax[1, xx].legend(["signal", "envelope", "sync"])

        latency_warning = idx_labels[xx] + ": neg lats in: "
        if len(latency_flags) > 0:
            for flag in latency_flags:
                latency_warning = latency_warning + flag + ", "
            ax[0, xx].set_title(latency_warning[:-2])
        else:
            ax[0, xx].set_title(idx_labels[xx])

    if session.task_version and "templeton" in session.task_version:
        figtitle = (
            "Comp tone vs. AM noise latency: " + session.id + " RFMapping (templeton)"
        )
    else:
        figtitle = "Comp tone vs. AM noise latency: " + session.id + " RFMapping (DR)"

    fig.suptitle(figtitle)
    fig.tight_layout()

    return fig



"""
getting Optotagging trials table requires:
- one or more hdf5 files with trial/stim data, called 'OptoTagging_*.hdf5'
- frame display times from sync, to use in place of frametimes in hdf5 file
- latency estimate for each stim presentation, to be added to frame
display times to get stim onset times
"""

from __future__ import annotations

from collections.abc import Iterable

import DynamicRoutingTask.TaskUtils
import numpy as np
import numpy.typing as npt

import npc_sessions.utils as utils
from npc_sessions.trials.TaskControl import TaskControl


class OptoTagging(TaskControl):
    """
    >>> stim = 's3://aind-ephys-data/ecephys_662892_2023-08-21_12-43-45/behavior/OptoTagging_662892_20230821_125915.hdf5'
    >>> sync = 's3://aind-ephys-data/ecephys_662892_2023-08-21_12-43-45/behavior/20230821T124345.h5'
    >>> trials = OptoTagging(stim, sync)
    >>> assert not trials._df.is_empty()
    """

    def __init__(
        self,
        hdf5: utils.StimPathOrDataset,
        sync: utils.SyncPathOrDataset,
        ephys_recording_dirs: Iterable[utils.PathLike] | None = None,
        **kwargs,
    ) -> None:
        if sync is None:
            raise ValueError(
                f"sync data is required for {self.__class__.__name__} trials table"
            )
        self._ephys_recording_dirs = ephys_recording_dirs
        super().__init__(
            hdf5, sync, ephys_recording_dirs=ephys_recording_dirs, **kwargs
        )

    @utils.cached_property
    def _stim_recordings(self) -> tuple[utils.StimRecording, ...]:
        recordings = utils.get_stim_latencies_from_sync(
            self._hdf5, self._sync, waveform_type="opto"
        )
        assert (
            None not in recordings
        ), f"{recordings.count(None) = } encountered: expected a recording of stim onset for every trial"
        return tuple(_ for _ in recordings if _ is not None)  # for mypy
        # TODO check this works for all older sessions

    @utils.cached_property
    def trial_index(self) -> npt.NDArray[np.int32]:
        """0-indexed"""
        return np.arange(len(self._hdf5["trialOptoOnsetFrame"]))

    @utils.cached_property
    def _inter_trial_interval(self) -> float:
        return self._hdf5["optoInterval"][()] / self._hdf5["frameRate"][()]

    @utils.cached_property
    def start_time(self) -> npt.NDArray[np.float64]:
        return self.stim_start_time - min(0.2, 0.5 * self._inter_trial_interval)

    @utils.cached_property
    def stop_time(self) -> npt.NDArray[np.float64]:
        return self.stim_stop_time + min(0.2, 0.5 * self._inter_trial_interval)

    @utils.cached_property
    def stim_start_time(self) -> npt.NDArray[np.float64]:
        return np.array([rec.onset_time_on_sync for rec in self._stim_recordings])[
            self.trial_index
        ]

    @utils.cached_property
    def stim_stop_time(self) -> npt.NDArray[np.float64]:
        return np.array([rec.offset_time_on_sync for rec in self._stim_recordings])[
            self.trial_index
        ]

    @utils.cached_property
    def stim_name(self) -> npt.NDArray[np.str_]:
        return np.array([recording.name for recording in self._stim_recordings])[
            self.trial_index
        ]

    @utils.cached_property
    def duration(self) -> npt.NDArray[np.float64]:
        return self._hdf5["trialOptoDur"][self.trial_index]

    @utils.cached_property
    def location(self) -> npt.NDArray[np.str_]:
        if all(str(v).upper() in "ABCDEF" for v in self._location):
            return np.array(
                [f"probe{str(v).upper()}" for v in self._location], dtype=str
            )
        return self._location

    @utils.cached_property
    def _bregma_xy(self) -> tuple[tuple[np.float64, np.float64], ...]:
        bregma = self._hdf5.get("optoBregma", None) or self._hdf5.get("bregmaXY", None)
        galvo = self._hdf5["galvoVoltage"][()]
        trial_voltages = self._hdf5["trialGalvoVoltage"]
        return tuple(
            tuple(bregma[np.all(galvo == v, axis=1)][0]) for v in trial_voltages
        )

    @utils.cached_property
    def bregma_x(self) -> npt.NDArray[np.float64]:
        return np.array([bregma[0] for bregma in self._bregma_xy])[self.trial_index]

    @utils.cached_property
    def bregma_y(self) -> npt.NDArray[np.float64]:
        return np.array([bregma[1] for bregma in self._bregma_xy])[self.trial_index]

    @utils.cached_property
    def _location(self) -> npt.NDArray[np.str_]:
        if trialOptoLabel := self._hdf5.get("trialOptoLabel", None):
            return np.array(trialOptoLabel.asstr()[self.trial_index], dtype=str)
        if optoTaggingLocs := self._hdf5.get("optoTaggingLocs"):
            label = optoTaggingLocs["label"].asstr()[()]
            xy = np.array(list(zip(optoTaggingLocs["X"], optoTaggingLocs["Y"])))
            return np.array(
                [label[np.all(xy == v, axis=1)][0] for v in self._bregma_xy], dtype=str
            )[self.trial_index]
        raise ValueError("No known optotagging location data found")

    @utils.cached_property
    def power(self) -> npt.NDArray[np.float64]:
        calibration_data = self._hdf5["optoPowerCalibrationData"]
        trial_voltages = self._hdf5["trialOptoVoltage"][self.trial_index]
        if "poly coefficients" in calibration_data:
            return DynamicRoutingTask.TaskUtils.voltsToPower(
                calibration_data,
                trial_voltages,
            )
        return np.where(~np.isnan(trial_voltages), self._hdf5["optoPower"], np.nan)
        # return trial_voltages * calibration_data['slope'] + calibration_data['intercept']


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )
